# import random
#
# import numpy as np
#
# import torch
# import torch.nn as nn
# from torch.autograd import Variable
# import torch.nn.functional as F
#
# from .attention import Attention
# from .baseRNN import BaseRNN
#
# if torch.cuda.is_available():
#     import torch.cuda as device
# else:
#     import torch as device
#
#
# class DecoderRNN(BaseRNN):
#     r"""
#     Provides functionality for decoding in a seq2seq framework, with an option for attention.
#     Args:
#         vocab_size (int): size of the vocabulary
#         max_len (int): a maximum allowed length for the sequence to be processed
#         hidden_size (int): the number of features in the hidden state `h`
#         sos_id (int): index of the start of sentence symbol
#         eos_id (int): index of the end of sentence symbol
#         n_layers (int, optional): number of recurrent layers (default: 1)
#         rnn_cell (str, optional): type of RNN cell (default: gru)
#         bidirectional (bool, optional): if the encoder is bidirectional (default False)
#         input_dropout_p (float, optional): dropout probability for the input sequence (default: 0)
#         dropout_p (float, optional): dropout probability for the output sequence (default: 0)
#         use_attention(bool, optional): flag indication whether to use attention mechanism or not (default: false)
#     Attributes:
#         KEY_ATTN_SCORE (str): key used to indicate attention weights in `ret_dict`
#         KEY_LENGTH (str): key used to indicate a list representing lengths of output sequences in `ret_dict`
#         KEY_SEQUENCE (str): key used to indicate a list of sequences in `ret_dict`
#     Inputs: inputs, encoder_hidden, encoder_outputs, function, teacher_forcing_ratio
#         - **inputs** (batch, seq_len, input_size): list of sequences, whose length is the batch size and within which
#           each sequence is a list of token IDs.  It is used for teacher forcing when provided. (default `None`)
#         - **encoder_hidden** (num_layers * num_directions, batch_size, hidden_size): tensor containing the features in the
#           hidden state `h` of encoder. Used as the initial hidden state of the decoder. (default `None`)
#         - **encoder_outputs** (batch, seq_len, hidden_size): tensor with containing the outputs of the encoder.
#           Used for attention mechanism (default is `None`).
#         - **function** (torch.nn.Module): A function used to generate symbols from RNN hidden state
#           (default is `torch.nn.functional.log_softmax`).
#         - **teacher_forcing_ratio** (float): The probability that teacher forcing will be used. A random number is
#           drawn uniformly from 0-1 for every decoding token, and if the sample is smaller than the given value,
#           teacher forcing would be used (default is 0).
#     Outputs: decoder_outputs, decoder_hidden, ret_dict
#         - **decoder_outputs** (seq_len, batch, vocab_size): list of tensors with size (batch_size, vocab_size) containing
#           the outputs of the decoding function.
#         - **decoder_hidden** (num_layers * num_directions, batch, hidden_size): tensor containing the last hidden
#           state of the decoder.
#         - **ret_dict**: dictionary containing additional information as follows {*KEY_LENGTH* : list of integers
#           representing lengths of output sequences, *KEY_SEQUENCE* : list of sequences, where each sequence is a list of
#           predicted token IDs }.
#     """
#
#     KEY_ATTN_SCORE = 'attention_score'
#     KEY_LENGTH = 'length'
#     KEY_SEQUENCE = 'sequence'
#
#     def __init__(self, vocab_size, max_len, hidden_size,
#             sos_id, eos_id,
#             n_layers=1, rnn_cell='gru', bidirectional=False,
#             input_dropout_p=0, dropout_p=0, use_attention=False):
#         super(DecoderRNN, self).__init__(vocab_size, max_len, hidden_size,
#                 input_dropout_p, dropout_p,
#                 n_layers, rnn_cell)
#
#         self.bidirectional_encoder = bidirectional
#         self.rnn = self.rnn_cell(hidden_size, hidden_size, n_layers, batch_first=True, dropout=dropout_p)
#
#         self.output_size = vocab_size
#         self.max_length = max_len
#         self.use_attention = use_attention
#         self.eos_id = eos_id
#         self.sos_id = sos_id
#
#         self.init_input = None
#
#         self.embedding = nn.Embedding(self.output_size, self.hidden_size)
#         if use_attention:
#             self.attention = Attention(self.hidden_size)
#
#         self.out = nn.Linear(self.hidden_size, self.output_size)
#
#     def forward_step(self, input_var, hidden, encoder_outputs, function):
#         batch_size = input_var.size(0)
#         output_size = input_var.size(1)
#         embedded = self.embedding(input_var)
#         embedded = self.input_dropout(embedded)
#
#         output, hidden = self.rnn(embedded, hidden)
#
#         attn = None
#         if self.use_attention:
#             output, attn = self.attention(output, encoder_outputs)
#
#         predicted_softmax = function(self.out(output.contiguous().view(-1, self.hidden_size)), dim=1).view(batch_size, output_size, -1)
#         return predicted_softmax, hidden, attn
#
#     def forward(self, inputs=None, encoder_hidden=None, encoder_outputs=None,
#                     function=F.log_softmax, teacher_forcing_ratio=0):
#         ret_dict = dict()
#         if self.use_attention:
#             ret_dict[DecoderRNN.KEY_ATTN_SCORE] = list()
#
#         inputs, batch_size, max_length = self._validate_args(inputs, encoder_hidden, encoder_outputs,
#                                                              function, teacher_forcing_ratio)
#         decoder_hidden = self._init_state(encoder_hidden)
#
#         use_teacher_forcing = True if random.random() < teacher_forcing_ratio else False
#
#         decoder_outputs = []
#         sequence_symbols = []
#         lengths = np.array([max_length] * batch_size)
#
#         def decode(step, step_output, step_attn):
#             decoder_outputs.append(step_output)
#             if self.use_attention:
#                 ret_dict[DecoderRNN.KEY_ATTN_SCORE].append(step_attn)
#             symbols = decoder_outputs[-1].topk(1)[1]
#             sequence_symbols.append(symbols)
#
#             eos_batches = symbols.data.eq(self.eos_id)
#             if eos_batches.dim() > 0:
#                 eos_batches = eos_batches.cpu().view(-1).numpy()
#                 update_idx = ((lengths > step) & eos_batches) != 0
#                 lengths[update_idx] = len(sequence_symbols)
#             return symbols
#
#         # Manual unrolling is used to support random teacher forcing.
#         # If teacher_forcing_ratio is True or False instead of a probability, the unrolling can be done in graph
#         if use_teacher_forcing:
#             decoder_input = inputs[:, :-1]
#             decoder_output, decoder_hidden, attn = self.forward_step(decoder_input, decoder_hidden, encoder_outputs,
#                                                                      function=function)
#
#             for di in range(decoder_output.size(1)):
#                 step_output = decoder_output[:, di, :]
#                 if attn is not None:
#                     step_attn = attn[:, di, :]
#                 else:
#                     step_attn = None
#                 decode(di, step_output, step_attn)
#         else:
#             decoder_input = inputs[:, 0].unsqueeze(1)
#             for di in range(max_length):
#                 decoder_output, decoder_hidden, step_attn = self.forward_step(decoder_input, decoder_hidden, encoder_outputs,
#                                                                          function=function)
#                 step_output = decoder_output.squeeze(1)
#                 symbols = decode(di, step_output, step_attn)
#                 decoder_input = symbols
#
#         ret_dict[DecoderRNN.KEY_SEQUENCE] = sequence_symbols
#         ret_dict[DecoderRNN.KEY_LENGTH] = lengths.tolist()
#
#         return decoder_outputs, decoder_hidden, ret_dict
#
#     def _init_state(self, encoder_hidden):
#         """ Initialize the encoder hidden state. """
#         if encoder_hidden is None:
#             return None
#         if isinstance(encoder_hidden, tuple):
#             encoder_hidden = tuple([self._cat_directions(h) for h in encoder_hidden])
#         else:
#             encoder_hidden = self._cat_directions(encoder_hidden)
#         return encoder_hidden
#
#     def _cat_directions(self, h):
#         """ If the encoder is bidirectional, do the following transformation.
#             (#directions * #layers, #batch, hidden_size) -> (#layers, #batch, #directions * hidden_size)
#         """
#         if self.bidirectional_encoder:
#             h = torch.cat([h[0:h.size(0):2], h[1:h.size(0):2]], 2)
#         return h
#
#     def _validate_args(self, inputs, encoder_hidden, encoder_outputs, function, teacher_forcing_ratio):
#         if self.use_attention:
#             if encoder_outputs is None:
#                 raise ValueError("Argument encoder_outputs cannot be None when attention is used.")
#
#         # inference batch size
#         if inputs is None and encoder_hidden is None:
#             batch_size = 1
#         else:
#             if inputs is not None:
#                 batch_size = inputs.size(0)
#             else:
#                 if self.rnn_cell is nn.LSTM:
#                     batch_size = encoder_hidden[0].size(1)
#                 elif self.rnn_cell is nn.GRU:
#                     batch_size = encoder_hidden.size(1)
#
#         # set default input and max decoding length
#         if inputs is None:
#             if teacher_forcing_ratio > 0:
#                 raise ValueError("Teacher forcing has to be disabled (set 0) when no inputs is provided.")
#             inputs = torch.LongTensor([self.sos_id] * batch_size).view(batch_size, 1)
#             if torch.cuda.is_available():
#                 inputs = inputs.cuda()
#             max_length = self.max_length
#         else:
#             max_length = inputs.size(1) - 1 # minus the start of sequence symbol
#
#         return inputs, batch_size, max_length


#! /usr/local/env python3
"""Pytorch visualization code modified from Chad Jensen's implementation
(https://discuss.pytorch.org/t/lstm-for-sequence-prediction/22021/3)."""
try:
    import torch
except ImportError:
    raise ImportError("pytorch is not installed. Install it via pytorch.org")
import traja
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
import numpy as np
import matplotlib.pyplot as plt
import logging

plt.style.use('dark_background')

nb_steps = 10

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

class LSTM(nn.Module):

    def __init__(self):
        super(LSTM, self).__init__()
        self.lstm = nn.LSTM(2, 100)
        self.head = nn.Linear(100, 2)

    def forward(self, x):
        outputs, states = self.lstm(x)
        outputs = outputs.reshape(x.shape[0] * x.shape[1], -1)
        pred = self.head(outputs)

        return pred


class TrajectoryLSTM():

    def __init__(self, xy, nb_steps=10, epochs=1000, batch_size=1, criterion=nn.MSELoss()):
        fig, ax = plt.subplots(2, 1)
        self.fig = fig
        self.ax = ax
        self.xy = xy
        self.nb_steps = nb_steps
        self.epochs = epochs
        self.batch_size = batch_size
        self.criterion = criterion
        self.rnn = LSTM()

    def load_batch(self, batch_size=32):
        t_1_b = np.zeros((nb_steps, batch_size, 2))
        t_b = np.zeros((nb_steps * batch_size, 2))

        inds = np.random.randint(0, len(self.xy) - nb_steps, (batch_size))
        for i, ind in enumerate(inds):
            t_1_b[:, i] = self.xy[ind:ind + nb_steps]
            t_b[i * nb_steps:(i + 1) * nb_steps] = self.xy[ind + 1:ind + nb_steps + 1]
        return torch.from_numpy(t_1_b).float(), torch.from_numpy(t_b).float()

    def train(self):
        mean_loss = 0.
        for epoch in range(1, self.epochs + 1):
            t_1_b, t_b = self.load_batch(self.batch_size)

            def closure():
                global loss
                optimizer.zero_grad()
                pred = self.rnn(t_1_b)
                shaped_pred = pred.reshape(-1, 2)
                loss = self.criterion(abs(shaped_pred), abs(t_b))
                loss.backward()

                return loss

            optimizer = optim.Adam(self.rnn.parameters(), 1e-3)
            optimizer.step(closure)
            mean_loss += loss.item()

            if epoch % 100 == 0:
                print('Epoch: {} | Loss: {:.6f}'.format(epoch, mean_loss))
                mean_loss = 0

    def savefig(self, filepath):
        self.fig.savefig(filepath)

    def _plot(self):
        t_1_b, t_b = self.load_batch(1)
        pred = self.rnn(t_1_b).detach().numpy().reshape(-1, 2)

        real = t_1_b.numpy().reshape(-1, 2)
        x, y = self.xy.T
        self.ax[0].plot(x, y, label='Real')
        self.ax[0].plot(real[:, 0], real[:, 1], label='Real batch')
        self.ax[0].plot(pred[:, 0], pred[:, 1], label='Pred')

        self.ax[1].scatter(real[:, 0], real[:, 1], label='Real')
        self.ax[1].scatter(pred[:, 0], pred[:, 1], label='Pred')

        for a in self.ax:
            a.legend()

    def plot(self, interactive=False, as_generator=False):
        if interactive and (plt.get_backend() == 'agg'):
            logging.ERROR("Not able to use interactive plotting in mpl `agg` mode.")
            interactive = False
        if as_generator:
            while True:
                self._plot()
                yield self.fig
        elif interactive:
            while True:
                for a in self.ax:
                    a.clear()
                self._plot()
                plt.pause(0.1)
                plt.show()
                input("Press [enter] to continue.")

        else:
            self._plot()
            return self.fig


#! /usr/local/env python3
try:
    import rpy2
except ImportError:
    raise ImportError("Python package rpy2 is required for this. Install it with"
                      ""
                      "pip install rpy2"
                      "")
import rpy2.robjects.packages as rpackages
import pandas as pd
try:
    import rpy2.robjects.pandas2ri as rpandas
except ModuleNotFoundError as e:
    raise ModuleNotFoundError(e + "\n Install tzlocal with `pip install tzlocal`.")
from rpy2.robjects.packages import importr

rpandas.activate()

ADEHABITAT_INSTALLED=False

def import_adehabitat():
    global ADEHABITAT_INSTALLED
    if not ADEHABITAT_INSTALLED:
        utils = rpackages.importr('utils')
        utils.chooseCRANmirror(ind=1)
        utils.install_packages('adehabitatLT')
        ADEHABITAT_INSTALLED = True
    adehabitat = importr('adehabitatLT')
    return adehabitat

def plot_ltraj(ltraj, id=1):
    """Plot `ltraj` using R method."""
    adehabitat = import_adehabitat()
    adehabitat.plot_ltraj(ltraj, id=1)


def to_ltraj(trj, id=1, typeII=False):
    """Convert trajectory to R `ltraj` object.

    Args:
        trajectory (:class:`~traja.main.TrajaDataFrame`): trajectory
        id (str, optional): id of animal/target
        typeII (bool):  contains datatime format time series (not yet supported)

    Returns:
        ltraj (:class:`rpy2.robjects.vectors.ListVector`): first index (ie, ltraj[0]) is pandas DataFrame including
                                                            columns 'x', 'y', 'date', 'dx', 'dy', 'dist', 'dt',
                                                            'R2n', 'abs.angle', 'rel.angle'

    .. doctest::

        >>> import traja.rutils as rutils # doctest:+ELLIPSIS
        >>>
        >>> df=traja.TrajaDataFrame({'x':range(5),'y':range(5)})
        >>> ltraj = rutils.to_ltraj(df)
        >>> print(ltraj[0])
        ...
           x  y  date          dx    ...              dt   R2n  abs.angle  rel.angle
        0  0  0     1           1    ...               1   0.0   0.785398        NaN
        1  1  1     2           1    ...               1   2.0   0.785398        0.0
        2  2  2     3           1    ...               1   8.0   0.785398        0.0
        3  3  3     4           1    ...               1  18.0   0.785398        0.0
        4  4  4     5 -2147483648    ...     -2147483648  32.0        NaN        NaN

        [5 rows x 10 columns]

    """
    adehabitat = import_adehabitat()

    df = trj[['x','y']]

    if typeII:
        raise NotImplementedError("datetime series not yet implemented for this method.")
        # FIXME: Add date converted from rpy2.robjects.POSIXct
        # date = None
        ltraj = adehabitat.as_ltraj(df, id=id, date=date, typeII=True) # Doesn't work
    else:
        ltraj = adehabitat.as_ltraj(df, id=id, typeII=False)
    return ltraj



#! /usr/local/env python3
import logging

import traja
import matplotlib.pyplot as plt
from matplotlib.path import Path
import matplotlib.patches as patches
import numpy as np
import pandas as pd

from traja import TrajaDataFrame
from pandas.core.dtypes.common import is_datetime_or_timedelta_dtype, is_datetime64_any_dtype, is_timedelta64_dtype

logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.ERROR)


def stylize_axes(ax):
    """Add top and right border to plot, set ticks."""
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)

    ax.xaxis.set_tick_params(top='off', direction='out', width=1)
    ax.yaxis.set_tick_params(right='off', direction='out', width=1)


def shift_xtick_labels(xtick_labels, first_index=None):
    for idx, x in enumerate(xtick_labels):
        label = x.get_text()
        xtick_labels[idx].set_text(str(int(label) + 1))
        if first_index is not None:
            xtick_labels[0] = first_index
    return xtick_labels


def sans_serif():
    """Convenience function for changing plot text to serif font."""
    plt.rc('font', family='serif')


def predict(xy, nb_steps=10, epochs=1000, batch_size=1, model='lstm'):
    """Method for training and visualizing LSTM with trajectory data."""
    if model is 'lstm':
        from traja.models.nn import TrajectoryLSTM
        TrajectoryLSTM(xy, nb_steps=nb_steps, epochs=epochs, batch_size=batch_size)

def plot(trj, accessor=None, n_coords: int = None, show_time=False, **kwargs):
    """Plot trajectory for single animal over period.

    Args:
      trj (:class:traja.TrajaDataFrame): trajectory

      n_coords (int): Number of coordinates to plot
      **kwargs: additional keyword arguments to :meth:`matplotlib.axes.Axes.scatter`

    Returns:
        ax (:class:`~matplotlib.collections.PathCollection`): Axes of plot

    """
    GRAY = '#999999'
    self = kwargs.get('self', {})
    if accessor:
        self = accessor
        kwargs = self._get_plot_args(**kwargs)
    else:
        kwargs = kwargs
    xlim = kwargs.pop('xlim', None)
    ylim = kwargs.pop('ylim', None)
    spatial_units = trj.__dict__.get('spatial_units', None)
    xlabel = kwargs.pop('xlabel', None) or f'x ({spatial_units})' if spatial_units else ''
    ylabel = kwargs.pop('ylabel', None) or f'y ({spatial_units})' if spatial_units else ''
    title = kwargs.pop('title', None)
    time_units = kwargs.pop('time_units', None)
    fps = kwargs.pop('fps', None)
    figsize = kwargs.pop('figsize', None)

    start, end = None, None
    coords = trj[['x', 'y']]
    time_col = traja.trajectory._get_time_col(trj)
    if time_col is 'index':
        is_datetime = True
    else:
        is_datetime = is_datetime_or_timedelta_dtype(trj[time_col]) if time_col else False

    if n_coords is None:
        # Plot all coords
        start, end = 0, len(coords)
        verts = coords.iloc[:end].values
    else:
        # Plot first `n_coords`
        start, end = 0, n_coords
        verts = coords.iloc[:n_coords].values

    n_coords = len(verts)

    codes = [Path.MOVETO] + [Path.LINETO] * (len(verts) - 1)
    path = Path(verts, codes)

    fig, ax = plt.subplots(figsize=figsize)
    fig.canvas.draw()
    patch = patches.PathPatch(path, edgecolor=GRAY, facecolor='none', lw=3, alpha=0.3)
    ax.add_patch(patch)

    xs, ys = zip(*verts)

    if time_col is 'index':
        # DatetimeIndex determines color
        colors = [ind for ind, x in enumerate(trj.index[:n_coords])]
    elif time_col and time_col is not 'index':
        # `time_col` determines color
        colors = [ind for ind, x in enumerate(trj[time_col].iloc[:n_coords])]
    else:
        # Frame count determines color
        colors = trj.index[:n_coords]

    if time_col:
        # TODO: Calculate fps if not in datetime
        vmin = min(colors)
        vmax = max(colors)
        if is_datetime:
            # Show timestamps without units
            time_units = ''
    else:
        # Index/frame count is our only reference
        vmin = trj.index[0]
        vmax = trj.index[n_coords - 1]
        if not show_time:
            time_units = ''
    label = f"Time ({time_units})" if time_units else ""

    sc = ax.scatter(xs, ys, c=colors, s=25, cmap=plt.cm.viridis, alpha=0.7, vmin=vmin, vmax=vmax)

    if xlim is not None:
        ax.set_xlim(xlim)
    else:
        ax.set_xlim((coords.x.min(), coords.x.max()))
    if ylim is not None:
        ax.set_ylim(ylim)
    else:
        ax.set_ylim((coords.y.min(), coords.y.max()))

    if kwargs.pop('invert_yaxis', None):
        plt.gca().invert_yaxis()

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(title)
    ax.set_aspect('equal')

    # Number of color bar ticks
    # FIXME: Implement customizable
    CBAR_TICKS = 10 if n_coords > 20 else n_coords
    indices = np.linspace(0, n_coords - 1, CBAR_TICKS, endpoint=True, dtype=int)
    cbar = plt.colorbar(sc, fraction=0.046, pad=0.04, orientation='vertical', label=label)
    if time_col is 'index':
        if is_datetime64_any_dtype(trj.index):
            cbar_labels = trj.index[indices].strftime("%Y-%m-%d %H:%M:%S").values.astype(str)
        elif is_timedelta64_dtype(trj.index):
            if time_units in ('s', '', None):
                cbar_labels = [round(x,2) for x in trj.index[indices].total_seconds()]
            else:
                logging.ERROR("Time unit {} not yet implemented".format(time_units))
        else:
            raise NotImplementedError("Indexing on {} is not yet implemented".format(type(trj.index)))
    elif time_col and is_timedelta64_dtype(trj[time_col]):
        cbar_labels = trj[time_col].iloc[indices].dt.total_seconds().values.astype(str)
    elif time_col and is_datetime:
        cbar_labels = trj[time_col].iloc[indices].dt.strftime("%Y-%m-%d %H:%M:%S").values.astype(str)
    else:
        # Convert frames to time
        cbar_labels = trj.index[indices].values
        if fps is not None and fps > 0 and fps is not 1 and show_time:
            cbar_labels = cbar_labels / fps
    cbar.set_ticks(indices)
    cbar.set_ticklabels(cbar_labels)
    plt.tight_layout()
    plt.show()
    return ax


def trip_grid(trj, bins=16, log=False, spatial_units=None, normalize=False, hist_only=False):
    """Generate a heatmap of time spent by point-to-cell gridding.

    Args:
      bins (int, optional): Number of bins (Default value = 16)
      log (bool): log scale histogram (Default value = False)
      spatial_units (str): units for plotting
      normalize (bool): normalize histogram into density plot
      hist_only (bool): return histogram without plotting

    Returns:
        hist (:class:`numpy.ndarray`): 2D histogram as array
        image (:class:`matplotlib.collections.PathCollection`: image of histogram

    """
    # TODO: Add kde-based method for line-to-cell gridding
    df = trj[['x', 'y']].dropna()

    # Set aspect if `xlim` and `ylim` set.
    if 'xlim' in df.__dict__ and 'ylim' in df.__dict__ and isinstance(df.xlim,tuple): # TrajaDataFrame
        x0, x1 = df.xlim
        y0, y1 = df.ylim
    else:
        x0, x1 = (df.x.min(), df.x.max())
        y0, y1 = (df.y.min(), df.y.max())
    aspect = (y1 - y0) / (x1 - x0)
    x_edges = np.linspace(x0, x1, num=bins)
    y_edges = np.linspace(y0, y1, num=int(bins / aspect))

    x, y = zip(*df.values)
    # # TODO: Remove redundant histogram calculation
    hist, x_edges, y_edges = np.histogram2d(x, y, bins=(x_edges, y_edges), density=normalize)
    if log:
        hist = np.log(hist + np.e)
    if hist_only:
        return hist, None
    fig, ax = plt.subplots()
    image = plt.imshow(hist, interpolation='bilinear')
    # TODO: Adjust colorbar ytick_labels to correspond with time
    cbar = plt.colorbar(image, ax=ax)
    ax.set_xlabel("x{}".format(" (" + spatial_units + ")" if spatial_units else ""))
    ax.set_ylabel("y{}".format(" (" + spatial_units + ")" if spatial_units else ""))
    plt.title("Time spent{}".format(' (Logarithmic)' if log else ''))
    plt.tight_layout()
    plt.show()
    # TODO: Add method for most common locations in grid
    # peak_index = unravel_index(hist.argmax(), hist.shape)
    return hist, image


def read_file(filepath:str, id=None, xlim=None, ylim=None, spatial_units='m', fps=None, **kwargs):
    """Convenience method wrapping pandas `read_csv` and initializing metadata.

    Args:
      filepath (str): path to csv file with `x`, `y` and `time` (optional) columns
      **kwargs: Additional arguments for :meth:`pandas.read_csv`.

    Returns:
        traj_df (:class:`~traja.main.TrajaDataFrame`): Trajectory

    """

    date_parser = kwargs.pop('data_parser', None)

    # TODO: Set index to first column containing 'time'
    df_test = pd.read_csv(filepath, nrows=10, parse_dates=True, infer_datetime_format=True)

    # Strip whitespace
    whitespace_cols = [c for c in df_test if ' ' in df_test[c].name]
    stripped_cols = {c: lambda x: x.strip() for c in whitespace_cols}
    converters = {**stripped_cols, **kwargs.pop('converters', {})}

    # Downcast to float32 # TODO: Benchmark float32 vs float64 for very big datasets
    float_cols = [c for c in df_test if 'float' in df_test[c].dtype]
    float32_cols = {c: np.float32 for c in float_cols}

    # Convert string columns to categories
    string_cols = [c for c in df_test if df_test[c].dtype == str]
    category_cols = {c: 'category' for c in string_cols}
    dtype = {**float32_cols, **category_cols, **kwargs.pop('dtype', {})}

    time_cols = [col for col in df_test.columns if 'time' in col.lower()]

    if 'csv' in filepath:
        trj = pd.read_csv(filepath,
                          date_parser=date_parser,
                          infer_datetime_format=kwargs.pop('infer_datetime_format', True),
                          parse_dates=kwargs.pop('parse_dates', True),
                          converters=converters,
                          dtype=dtype,
                          **kwargs)
        if time_cols:
            time_col = time_cols[0]
            trj.rename(columns={time_col: 'time'})
        else:
            time = (trj.index) / fps
            trj['time'] = time
    else:
        # TODO: Implement for HDF5 and .npy files.
        raise NotImplementedError("Non-csv's not yet implemented")

    trj = TrajaDataFrame(trj)
    # Set meta properties of TrajaDataFrame
    trj.id = id
    trj.xlim = xlim
    trj.ylim = ylim
    trj.spatial_units = spatial_units
    trj.title = kwargs.get('title', None)
    trj.xlabel = kwargs.get('xlabel',None)
    trj.ylabel = kwargs.get('ylabel',None)
    trj.fps = fps
    return trj



# import torch.nn as nn
#
# from .baseRNN import BaseRNN
#
# class EncoderRNN(BaseRNN):
#     r"""
#     Applies a multi-layer RNN to an input sequence.
#     Args:
#         vocab_size (int): size of the vocabulary
#         max_len (int): a maximum allowed length for the sequence to be processed
#         hidden_size (int): the number of features in the hidden state `h`
#         input_dropout_p (float, optional): dropout probability for the input sequence (default: 0)
#         dropout_p (float, optional): dropout probability for the output sequence (default: 0)
#         n_layers (int, optional): number of recurrent layers (default: 1)
#         bidirectional (bool, optional): if True, becomes a bidirectional encodr (defulat False)
#         rnn_cell (str, optional): type of RNN cell (default: gru)
#         variable_lengths (bool, optional): if use variable length RNN (default: False)
#         embedding (torch.Tensor, optional): Pre-trained embedding.  The size of the tensor has to match
#             the size of the embedding parameter: (vocab_size, hidden_size).  The embedding layer would be initialized
#             with the tensor if provided (default: None).
#         update_embedding (bool, optional): If the embedding should be updated during training (default: False).
#     Inputs: inputs, input_lengths
#         - **inputs**: list of sequences, whose length is the batch size and within which each sequence is a list of token IDs.
#         - **input_lengths** (list of int, optional): list that contains the lengths of sequences
#             in the mini-batch, it must be provided when using variable length RNN (default: `None`)
#     Outputs: output, hidden
#         - **output** (batch, seq_len, hidden_size): tensor containing the encoded features of the input sequence
#         - **hidden** (num_layers * num_directions, batch, hidden_size): tensor containing the features in the hidden state `h`
#     Examples::
#          >>> encoder = EncoderRNN(input_vocab, max_seq_length, hidden_size)
#          >>> output, hidden = encoder(input)
#     """
#
#     def __init__(self, vocab_size, max_len, hidden_size,
#                  input_dropout_p=0, dropout_p=0,
#                  n_layers=1, bidirectional=False, rnn_cell='gru', variable_lengths=False,
#                  embedding=None, update_embedding=True):
#         super(EncoderRNN, self).__init__(vocab_size, max_len, hidden_size,
#                 input_dropout_p, dropout_p, n_layers, rnn_cell)
#
#         self.variable_lengths = variable_lengths
#         self.embedding = nn.Embedding(vocab_size, hidden_size)
#         if embedding is not None:
#             self.embedding.weight = nn.Parameter(embedding)
#         self.embedding.weight.requires_grad = update_embedding
#         self.rnn = self.rnn_cell(hidden_size, hidden_size, n_layers,
#                                  batch_first=True, bidirectional=bidirectional, dropout=dropout_p)
#
#     def forward(self, input_var, input_lengths=None):
#         """
#         Applies a multi-layer RNN to an input sequence.
#         Args:
#             input_var (batch, seq_len): tensor containing the features of the input sequence.
#             input_lengths (list of int, optional): A list that contains the lengths of sequences
#               in the mini-batch
#         Returns: output, hidden
#             - **output** (batch, seq_len, hidden_size): variable containing the encoded features of the input sequence
#             - **hidden** (num_layers * num_directions, batch, hidden_size): variable containing the features in the hidden state h
#         """
#         embedded = self.embedding(input_var)
#         embedded = self.input_dropout(embedded)
#         if self.variable_lengths:
#             embedded = nn.utils.rnn.pack_padded_sequence(embedded, input_lengths, batch_first=True)
#         output, hidden = self.rnn(embedded)
#         if self.variable_lengths:
#             output, _ = nn.utils.rnn.pad_packed_sequence(output, batch_first=True)
#         return output, hidden


#! /usr/local/env python3
import math
from typing import Callable

import traja
import numpy as np
import pandas as pd
import scipy

from traja import TrajaDataFrame
from pandas.core.dtypes.common import is_datetime_or_timedelta_dtype, is_datetime64_any_dtype, is_timedelta64_dtype
from scipy.spatial.distance import directed_hausdorff, euclidean


def smooth_sg(trj, w=None, p=3):
    """Savitzky-Golay filtering.

    Args:
      trj (:class:`~traja.main.TrajaDataFrame`): Trajectory
      w (int): window size (Default value = None)
      p (int): polynomial order (Default value = 3)

    Returns:
      trj: :class:`~traja.main.TrajaDataFrame`

    """
    if w is None:
        w = p + 3 - p % 2

    if (w % 2 != 1):
        raise Exception(f"Invalid smoothing parameter w ({w}): n must be odd")
    trj.x = scipy.signal.savgol_filter(trj.x, window_length=w, polyorder=p, axis=0)
    trj.y = scipy.signal.savgol_filter(trj.y, window_length=w, polyorder=p, axis=0)
    trj = fill_in_traj(trj)
    return trj


def angles(trj, lag=1, compass_direction=None):
    trj['angle'] = np.rad2deg(np.arccos(np.abs(trj['dx']) / trj['distance']))
    # Get heading from angle
    mask = (trj['dx'] > 0) & (trj['dy'] >= 0)
    trj.loc[mask, 'heading'] = trj['angle'][mask]
    mask = (trj['dx'] >= 0) & (trj['dy'] < 0)
    trj.loc[mask, 'heading'] = -trj['angle'][mask]
    mask = (trj['dx'] < 0) & (trj['dy'] <= 0)
    trj.loc[mask, 'heading'] = -(180 - trj['angle'][mask])
    mask = (trj['dx'] <= 0) & (trj['dy'] > 0)
    trj.loc[mask, 'heading'] = (180 - trj['angle'])[mask]
    trj['turn_angle'] = trj['heading'].diff()
    # Correction for 360-degree angle range
    trj.loc[trj.turn_angle >= 180, 'turn_angle'] -= 360
    trj.loc[trj.turn_angle < -180, 'turn_angle'] += 360


def step_lengths(trj):
    """Length of the steps of `trj`.

    Args:
      trj (:class:`~traja.main.TrajaDataFrame`): Trajectory

    Returns:

    """
    raise NotImplementedError()


def polar_to_z(r: float, theta: float):
    """Converts polar coordinates `z` and `theta` to complex number `z`.

    Args:
      r (float): step size
      theta (float): angle

    Returns:

    """
    return r * np.exp(1j * theta)


def cartesian_to_polar(xy):
    """Convert :class:`numpy.ndarray` `xy` to polar coordinates `r` and `theta`.

    Args:
      xy (:class:`numpy.ndarray`): x,y coordinates

    Returns:
        r, theta (tuple of float): step-length and angle

    """
    assert xy.ndim == 2, f"Dimensions are {xy.ndim}, expecting 2"
    x, y = np.split(xy, [-1], axis=1)
    x, y = np.squeeze(x), np.squeeze(y)
    r = math.sqrt(x * x + y * y)
    theta = math.atan2(y, x)
    return r, theta


def expected_sq_displacement(trj, n=None, eqn1=True, compass_direction=None):
    """Expected displacment.

    .. note::

        This method is experimental and needs testing.

    """
    # TODO: Fix and test implementation
    sl = step_lengths(trj)
    ta = angles(trj, compass_direction=compass_direction)
    l = np.mean(sl)
    l2 = np.mean(sl ^ 2)
    c = np.mean(np.cos(ta))
    s = np.mean(np.sin(ta))
    s2 = s ^ 2

    if eqn1:
        # Eqn 1
        alpha = np.arctan2(s, c)
        gamma = ((1 - c) ^ 2 - s2) * np.cos((n + 1) * alpha) - 2 * s * (1 - c) * np.sin((n + 1) * alpha)
        esd = n * l2 + 2 * l ^ 2 * ((c - c ^ 2 - s2) * n - c) / ((1 - c) ^ 2 + s2) + 2 * l ^ 2 * (
                (2 * s2 + (c + s2) ^ ((n + 1) / 2)) / ((1 - c) ^ 2 + s2) ^ 2) * gamma
        return abs(esd)
    else:
        # Eqn 2
        esd = n * l2 + 2 * l ^ 2 * c / (1 - c) * (n - (1 - c ^ n) / (1 - c))
        return esd


def traj_from_coords(track, x_col=1, y_col=2, time_col=None, fps=4, spatial_units='m', time_units='s'):
    # TODO: Convert to DataFrame if not already
    trj = track
    trj.traja.spatial_units = spatial_units
    trj.traja.time_units = time_units

    def rename(col, name):
        global trj
        if isinstance(col, int):
            trj.rename(columns={col: name})
        else:
            if col not in trj:
                raise Exception(f"Missing column {col}")
            trj.rename(columns={col: name})

    # Ensure column names are as expected
    rename(x_col, 'x')
    rename(y_col, 'y')
    if time_col is not None:
        rename(time_col, 'time')

    # Allocate times if they aren't already known
    if 'time' not in trj:
        if fps is None:
            raise Exception(("Cannot create a trajectory without times: either fps or a time column must be specified"))
        # Assign times to each frame, starting at 0
        trj['time'] = pd.Series(np.arange(0, len(trj) - 1) / fps)

    # Get displacement time for each coordinate, with the first point at time 0
    trj['dt'] = trj.time - trj.time.iloc[0]

    ...
    return trj


# TODO: Delete if unusable
# def traj(filepath, xlim=None, ylim=None, **kwargs):
#     df_test = pd.read_csv(filepath, nrows=100)
#     # Select first col with 'time_stamp' in name as index
#     time_stamp_cols = [x for x in df_test.columns if 'time_stamp' in x]
#     index_col = kwargs.pop('index_col', time_stamp_cols[0])
#
#     df = pd.read_csv(filepath,
#                      date_parser=kwargs.pop('date_parser',
#                                             lambda x: pd.datetime.strptime(x, '%Y-%m-%d %H:%M:%S:%f')),
#                      infer_datetime_format=kwargs.pop('infer_datetime_format', True),
#                      parse_dates=kwargs.pop('parse_dates', True),
#                      index_col=index_col,
#                      **kwargs)
#     if xlim is not None and isinstance(xlim, tuple):
#         df.traja.xlim = xlim
#     if ylim is not None and isinstance(ylim, tuple):
#         df.traja.ylim = ylim
#     return df


def distance(A: traja.TrajaDataFrame, B: traja.TrajaDataFrame, method='dtw'):
    """Calculate distance between two trajectories.

    Args:
        A (:class:`~traja.main.TrajaDataFrame`) : Trajectory 1
        B (:class:`~traja.main.TrajaDataFrame`) : Trajectory 2
        method (str): `dtw` for dynamic time warping, `hausdorff` for Hausdorff

    Returns:
        distance (str): Distance
    """
    if method is 'hausdorff':
        dist0 = directed_hausdorff(A, B)[0]
        dist1 = directed_hausdorff(B, A)[0]
        symmetric_dist = max(dist0, dist1)
        return symmetric_dist
    elif method is 'dtw':
        try:
            from fastdtw import fastdtw
        except ImportError:
            raise ImportError("""            
            fastdtw is not installed. Install it with: 
            pip install fastdtw.

            """)
        distance, path = fastdtw(A, B, dist=euclidean)
        return distance


def generate(n: int = 1000, random: bool = True, step_length: int = 2,
             angular_error_sd: float = 0.5,
             angular_error_dist: Callable = None,
             linear_error_sd: float = 0.2,
             linear_error_dist: Callable = None,
             fps: float = 50,
             spatial_units: str = 'm',
             seed: int = None,
             **kwargs):
    """Generates a trajectory.

    If `random` is `True`, the trajectory will
    be a correlated random walk/idiothetic directed walk (Kareiva & Shigesada,
    1983), corresponding to an animal navigating without a compass (Cheung,
    Zhang, Stricker, & Srinivasan, 2008). If `random` is `False`, it
    will be a directed walk/allothetic directed walk/oriented path, corresponding
    to an animal navigating with a compass (Cheung, Zhang, Stricker, &
    Srinivasan, 2007, 2008).

    By default, for both random and directed walks, errors are normally
    distributed, unbiased, and independent of each other, so are **simple
    directed walks** in the terminology of Cheung, Zhang, Stricker, & Srinivasan,
    (2008). This behaviour may be modified by specifying alternative values for
    the `angular_error_dist` and/or `linear_error_dist` parameters.

    The initial angle (for a random walk) or the intended direction (for a
    directed walk) is `0` radians. The starting position is `(0, 0)`.

    Args:
      n:  (Default value = 1000)
      random:  (Default value = True)
      step_length:  (Default value = 2)
      angular_error_sd:  (Default value = 0.5)
      angular_error_dist:  (Default value = None)
      linear_error_sd:  (Default value = 0.2)
      linear_error_dist:  (Default value = None)
      fps:  (Default value = 50)
      spatial_units:  (Default value = 'm')
      **kwargs:

    Returns:
        trj (:class:`traja.main.TrajaDataFrame`): Trajectory

    .. note::

        Based on Jim McLean's `trajr <https://github.com/JimMcL/trajr>`_, ported to Python by Justin Shenk.

        **Reference**: McLean, D. J., & Skowron Volponi, M. A. (2018). trajr: An R package for characterisation of animal
        trajectories. Ethology, 124(6), 440-448. https://doi.org/10.1111/eth.12739.

    """
    if seed is None:
        np.random.seed(0)
    if angular_error_dist is None:
        angular_error_dist = np.random.normal(loc=0., scale=angular_error_sd, size=n)
    if linear_error_dist is None:
        linear_error_dist = np.random.normal(loc=0., scale=linear_error_sd, size=n)
    angular_errors = angular_error_dist
    linear_errors = linear_error_dist
    step_lengths = step_length + linear_errors
    # Don't allow negative lengths
    step_lengths[step_lengths < 0] = 0
    steps = polar_to_z(step_lengths, angular_errors)

    if random:
        # Accumulate angular errors
        coords = np.zeros(n + 1, dtype=np.complex)
        angle = 0
        for i in range(n):
            angle += angular_errors[i]
            length = step_length + linear_errors[i]
            coords[i + 1] = coords[i] + polar_to_z(r=length, theta=angle)
    else:
        coords = np.append(complex(0), np.cumsum(steps))

    x = coords.real
    y = coords.imag

    df = traja.TrajaDataFrame(data={'x': x, 'y': y})
    if fps in (0, None):
        raise Exception("fps must be greater than 0")
    df.fps = fps
    time = df.index / fps
    df['time'] = time
    df.spatial_units = spatial_units
    for key, value in kwargs.items():
        df.__dict__[key] = value
    # Update metavars
    metavars = dict(angular_error_sd=angular_error_sd, linear_error_sd=linear_error_sd)
    df.__dict__.update(metavars)

    return df


def _resample_time(trj, step_time):
    if not is_datetime_or_timedelta_dtype(trj.index):
        raise Exception(f"{trj.index.dtype} is not datetime or timedelta.")
    return trj.resample(step_time).agg({'x': np.mean, 'y': np.mean})


def resample_time(trj, step_time, new_fps=None):
    """Resample trajectory to consistent `step_time` intervals.

    Args:
        trj (:class:`~traja.TrajaDataFrame`): trajectory
        step_time (str): step time interval (eg, '1s')
        new_fps (bool, optional): new fps

    Results:
        trj (:class:`~traja.TrajaDataFrame`): trajectory


    .. doctest::

        >>> from traja import generate
        >>> from traja.utils import resample_time
        >>> df = generate()
        >>> resampled = resample_time(df, '2s')
        >>> resampled.head()
              time          x          y
        0 00:00:00  14.555071 -26.482614
        1 00:00:02  -3.582797  -6.491297
        2 00:00:04  -4.299709  26.937443
        3 00:00:06 -25.337042  42.131848
        4 00:00:08  33.069915  32.780830



    """
    time_col = _get_time_col(trj)
    if time_col is 'index' and is_datetime64_any_dtype(trj.index):
        _trj = _resample_time(trj, step_time)
    elif time_col is 'index' and is_timedelta64_dtype(trj.index):
        _trj = _resample_time(trj, step_time)
    elif time_col:
        if isinstance(step_time, str):
            try:
                if '.' in step_time:
                    raise NotImplementedError("Fractional step time not implemented.")
            except Exception:
                raise NotImplementedError(f"Inferring from time format {step_time} not yet implemented.")
        _trj = trj.set_index(time_col)
        _trj.index = pd.to_timedelta(_trj.index, unit='s')
        _trj = _resample_time(_trj, step_time)
        _trj.reset_index(inplace=True)
    else:
        raise NotImplementedError(f"Time column ({time_col}) not of expected data type.")
    return _trj


def rotate(df, angle=0, origin=None):
    """Rotate a trajectory `angle` in radians.

    Args:
        trj: Traja.DataFrame
        angle

    Returns:
        trj: Traja.DataFrame

    .. note::

        Based on Lyle Scott's `implementation <https://gist.github.com/LyleScott/e36e08bfb23b1f87af68c9051f985302>`_.

    """
    trj = df.copy()
    # Calculate current orientation
    if isinstance(trj, traja.TrajaDataFrame):
        xy = df.traja.xy
    elif isinstance(trj, pd.DataFrame):
        trj = df[['x', 'y']]

    x, y = np.split(xy, [-1], axis=1)
    if origin is None:
        # Assume middle of x and y is origin
        origin = ((x.max() - x.min()) / 2, (y.max() - y.min()) / 2)

    offset_x, offset_y = origin
    new_coords = []

    for x, y in xy:
        adjusted_x = (x - offset_x)
        adjusted_y = (y - offset_y)
        cos_rad = math.cos(angle)
        sin_rad = math.sin(angle)
        qx = offset_x + cos_rad * adjusted_x + sin_rad * adjusted_y
        qy = offset_y + -sin_rad * adjusted_x + cos_rad * adjusted_y
        new_coords.append((qx, qy))

    new_xy = np.array(new_coords)
    x, y = np.split(new_xy, [-1], axis=1)
    trj['x'] = x
    trj['y'] = y
    return trj


def from_df(df):
    """Convenience function for converting a :class:`pandas DataFrame<pandas.dataframe.DataFrame>` into a :class:`traja.main.TrajaDataFrame`.

    Args:
      df (:class:`pandas.DataFrame`): Trajectory as pandas `DataFrame`

    Returns:
      traj_df (:class:`~traja.main.TrajaDataFrame`): Trajectory

    .. doctest::

        >>> df = pd.DataFrame({'x':[0,1,2],'y':[1,2,3]})
        >>> traja.from_df(df)
           x  y
        0  0  1
        1  1  2
        2  2  3

    """
    traj_df = TrajaDataFrame(df)
    # Initialize metadata
    for var in traj_df._metadata:
        if not hasattr(traj_df, var):
            traj_df.__dict__[var] = None
    return traj_df


def from_xy(xy: np.ndarray):
    """Convenience function for initializing :class:`TrajaDataFrame<traja.main.TrajaDataFrame>` with x,y coordinates.

    Args:
        xy (:class:`numpy.ndarray`): x,y coordinates

    Returns:
        traj_df (:class:`~traja.main.TrajaDataFrame`): Trajectory as dataframe

    .. doctest::

        >>> import numpy as np
        >>> xy = np.array([[0,1],[1,2],[2,3]])
        >>> traja.from_xy(xy)
           x  y
        0  0  1
        1  1  2
        2  2  3

    """
    df = traja.TrajaDataFrame.from_records(xy, columns=['x', 'y'])
    return df


def fill_in_traj(trj):
    # FIXME: Implement
    return trj


def _get_time_col(trj):
    # Check if saved in metadata
    time_col = trj.__dict__.get('time_col', None)
    if time_col:
        return time_col
    # Check if index is datetime
    if is_datetime_or_timedelta_dtype(trj.index):
        return 'index'
    # Check if any column contains 'time'
    time_cols = [col for col in trj if 'time' in col.lower()]
    if time_cols:
        # Try first column
        time_col = time_cols[0]
        if is_datetime_or_timedelta_dtype(trj[time_col]):
            return time_col
        else:
            # Time column is float, etc. but not datetime64.
            # FIXME: Add conditional return, etc.
            return time_col
    else:
        # No time column found
        return None



#! /usr/local/env python3
import argparse
import glob
import logging
import os
import sys
from collections import OrderedDict

import traja
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.dates import date2num, num2date, DateFormatter
from mpl_toolkits.axes_grid1 import make_axes_locatable, axes_size
from pandas.api.types import is_numeric_dtype, is_datetime64_any_dtype

from scipy.spatial.distance import directed_hausdorff, euclidean
from numpy import unravel_index
from shapely.geometry import shape
from traja.trajectory import polar_to_z

logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.ERROR)


@pd.api.extensions.register_dataframe_accessor("traja")
class TrajaAccessor(object):
    """Accessor for pandas DataFrame with trajectory-specific numerical and analytical functions."""

    def __init__(self, pandas_obj):
        self._trj = pandas_obj

    def _strip(self, text):
        try:
            return text.strip()
        except AttributeError:
            return pd.to_numeric(text, errors='coerce')

    def set(self, **kwargs):
        for key, value in kwargs:
            try:
                self._trj.__dict__[key] = value
            except Exception as e:
                raise Exception(f"Exception {e} assigning df.{key} to {value}")

    @property
    def night(self, begin='19:00', end='7:00'):
        """Returns trajectory indices for time between `begin` and `end`.

        Args:
          begin (str):  (Default value = '19:00')
          end (str):  (Default value = '7:00')

        Returns:
          trj (:class:`~traja.main.TrajaDataFrame`): Trajectory during night.

        """
        return self.between(begin, end)

    @property
    def day(self, begin='7:00', end='19:00'):
        """Get day between `begin` and `end`.

        Args:
          begin (str):  (Default value = '7:00')
          end (str):  (Default value = '19:00')

        Returns:
          trj (:class:`~traja.main.TrajaDataFrame`): Trajectory during day.

        """
        return self.between(begin, end)

    def set(self, **kwargs):
        for key, value in kwargs.items():
            try:
                self.__setattr__(key, value)
            except Exception as e:
                logging.ERROR(f"Cannot set {key} to {value}")

    def _get_plot_args(self, **kwargs):
        for var in self._trj._metadata:
            # Update global meta variables
            # TODO: Replace with elegant solution
            if var not in kwargs:
                # Metadata variable not specified in argument
                if var in self._trj.__dict__:
                    kwargs[var] = self._trj.__dict__[var]
        return kwargs

    def _get_time_col(self):
        """Search for time column in trajectory.

        Args:

        Returns:
           time_col (str or None): name of time column, 'index' or None

        """
        return traja.trajectory._get_time_col(self._trj)

    def between(self, begin, end):
        """Return trajectory between `begin` and end` if `time` column is `datetime64`.

        Args:
          begin (str): Beginning of time slice.
          end (str): End of time slice.

        Returns:
          trj (:class:`~traja.main.TrajaDataFrame`): Data frame between values.
          
        .. doctest ::

            >>> s = pd.to_datetime(pd.Series(['Jun 30 2000 12:00:01', 'Jun 30 2000 12:00:02', 'Jun 30 2000 12:00:03']))
            >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3],'time':s})
            >>> df.traja.between('12:00:00','12:00:01')
                             time  x  y
            0 2000-06-30 12:00:01  0  1

        """
        time_col = self._get_time_col()
        if time_col is 'index':
            return self._trj.between_time(begin, end)
        elif time_col and is_datetime64_any_dtype(self._trj[time_col]):
            # Backup index
            dt_index_col = self._trj.index.name
            # Set dt_index
            trj = self._trj.copy()
            trj.set_index(time_col, inplace=True)
            # Create slice of trajectory
            trj = trj.between_time(begin, end)
            # Restore index and return column
            if dt_index_col:
                trj.set_index(dt_index_col, inplace=True)
            else:
                trj.reset_index(inplace=True)
            return trj
        else:
            raise TypeError("Either time column or index must be datetime64")

    def resample_time(self, step_time):
        return traja.trajectory.resample_time(self._trj, step_time=step_time)

    def trip_grid(self, bins=16, log=False, spatial_units=None, normalize=False, hist_only=False):
        """Make a 2D histogram of trip.

        Args:
          bins (int, optional): Number of bins (Default value = 16)
          log (bool): log scale histogram (Default value = False)
          spatial_units (str): units for plotting
          normalize (bool): normalize histogram into density plot
          hist_only (bool): return histogram without plotting

        Returns:
            hist (:class:`numpy.ndarray`): 2D histogram as array
            image (:class:`matplotlib.collections.PathCollection`: image of histogram

        """
        hist, image = traja.plotting.trip_grid(self._trj,
                                               bins=bins,
                                               log=log,
                                               spatial_units=self._trj.spatial_units,
                                               normalize=normalize, hist_only=hist_only)
        return hist, image

    def plot(self, n_coords: int = None, show_time=False, **kwargs):
        """Plot trajectory for single animal over period.

        Args:
          n_coords (int): Number of coordinates to plot
          **kwargs: additional keyword arguments to :meth:`matplotlib.axes.Axes.scatter`

        Returns:
            ax (:class:`~matplotlib.collections.PathCollection`): Axes of plot
        """
        ax = traja.plotting.plot(trj=self._trj, accessor=self, n_coords=n_coords, show_time=show_time, **kwargs)
        return ax


    def _has_cols(self, cols: list):
        return set(cols).issubset(self._trj.columns)

    @property
    def xy(self, split=False):
        """Return numpy array of x,y coordinates.

        Args:
            split (bool): Split into seaprate x and y :class:`numpy.ndarrays`

        Returns:
          xy (:class:`numpy.ndarray`) -- x,y coordinates (separate if `split` is `True`)
          
        .. doctest::

            >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3]})
            >>> df.traja.xy
            array([[0, 1],
                   [1, 2],
                   [2, 3]])

        """
        if self._has_cols(['x', 'y']):
            xy = self._trj[['x', 'y']].values
            if split:
                xy = np.split(xy, [-1], axis=1)
            return xy
        else:
            raise Exception("'x' and 'y' are not in the dataframe.")

    def _check_has_time(self):
        """Check for presence of displacement time column."""
        if 'time' not in self._trj:
            raise Exception("Missing time information in trajectory.")

    def calc_derivatives(self, assign=False):
        """Calculate derivatives `displacement` and `displacement_time`.

        Args:
          assign (bool): Assign output to `TrajaDataFrame` (Default value = False)

        Returns:
          dict: Derivatives in dictionary.
          
        .. doctest::

            >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3],'time':[0., 0.2, 0.4]})
            >>> df.traja.calc_derivatives()
            OrderedDict([('displacement', 0         NaN
            1    1.414214
            2    1.414214
            dtype: float64), ('displacement_time', 0    0.0
            1    0.2
            2    0.4
            Name: time, dtype: float64)])

        """
        self._check_has_time()
        if not 'displacement' in self._trj:
            displacement = self.calc_displacement(assign=assign)
        else:
            displacement = self._trj.displacement

        displacement_time = self._trj.time - self._trj.time.iloc[0]

        derivs = OrderedDict(displacement=displacement, displacement_time=displacement_time)
        if assign:
            self._trj = self._trj.join(traja.TrajaDataFrame.from_records(derivs))
        return derivs

    def get_derivatives(self):
        """Get derivatives.

        Args:

        Returns:
          derivs (:class:`~collections.OrderedDict`) : Derivatives in dictionary.
          
        .. doctest::

            >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3],'time':[0.,0.2,0.4]})
            >>> df.traja.get_derivatives()
            OrderedDict([('displacement', 0         NaN
            1    1.414214
            2    1.414214
            dtype: float64), ('displacement_time', 0    0.0
            1    0.2
            2    0.4
            Name: time, dtype: float64), ('speed', 0         NaN
            1    7.071068
            2    7.071068
            dtype: float64), ('speed_times', 1    0.2
            2    0.4
            Name: speed_times, dtype: float64), ('acceleration', 0    NaN
            1    NaN
            2    0.0
            dtype: float64), ('acceleration_times', 2    0.4
            Name: accleration_times, dtype: float64)])

        """
        if not self._has_cols(['displacement', 'displacement_time']):
            derivs = self.calc_derivatives(assign=False)
            d = derivs['displacement']
            t = derivs['displacement_time']
        else:
            d = self._trj.displacement
            t = self._trj.displacement_time
            derivs = OrderedDict(displacement=d, displacement_time=t)
        v = d[1: len(d)] / t.diff()
        v.rename('speed')
        vt = t[1: len(t)].rename('speed_times')
        # Calculate linear acceleration
        a = v.diff() / vt.diff().rename('acceleration')
        at = vt[1: len(vt)].rename('accleration_times')
        data = OrderedDict(speed=v, speed_times=vt, acceleration=a, acceleration_times=at)
        derivs.update(data)
        return derivs

    @property
    def speed_intervals(self, faster_than=None, slower_than=None, interpolate_times=True):
        """Calculate speed time intervals.
        
        Returns a dictionary of time intervals where speed is slower and/or faster than specified values.

        Args:
          faster_than (float, optional): Minimum speed threshold. (Default value = None)
          slower_than (float or int, optional): Maximum speed threshold. (Default value = None)
          interpolate_times (bool, optional): Interpolate times between steps. (Default value = True)

        Returns:
          result (:class:`~collections.OrderedDict`) -- time intervals as dictionary.
          
        .. note::
          
            Implementation ported to Python, heavily inspired by Jim McLean's trajr package.

        """
        derivs = self.get_derivatives()

        if faster_than is not None:
            pass
        if slower_than is not None:
            pass

        # Calculate trajectory speeds
        speed = derivs.get('speed')
        times = derivs.get('speed_times')
        flags = np.full(len(speed), 1)

        if faster_than is not None:
            flags = flags & (speed > faster_than)
        if slower_than is not None:
            flags = flags & (speed < slower_than)

        changes = np.diff(flags)
        stop_frames = np.where(changes == -1)[0]
        start_frames = np.where(changes == 1)[0]

        # Handle situation where interval begins or ends outside of trajectory
        if len(start_frames) > 0 or len(stop_frames) > 0:
            # Assume interval started at beginning of trajectory, since we don't know what happened before that
            if len(stop_frames) > 0 and (len(start_frames) == 0 or stop_frames[0] < start_frames[0]):
                start_frames = np.append(1, start_frames)
            # Similarly, assume that interval can't extend past end of trajectory
            if len(stop_frames) == 0 or start_frames[len(start_frames) - 1] > stop_frames[len(stop_frames) - 1]:
                stop_frames = np.append(stop_frames, len(speed))

        stop_times = times[stop_frames]
        start_times = times[start_frames]

        if interpolate_times and len(start_frames) > 0:
            # TODO: Implement
            raise NotImplementedError()
            r = self.linear_interp_times(slower_than, faster_than, speed, times, start_frames, start_times)
            start_times = r[:, 0]
            stop_times = r[:, 1]

        durations = stop_times - start_times
        result = traja.TrajaDataFrame(OrderedDict(start_frame=start_frames,
                                           start_time=start_times,
                                           stop_frame=stop_frames,
                                           stop_time=stop_times,
                                           duration=durations))

        metadata = OrderedDict(slower_than=slower_than, faster_than=faster_than, derivs=derivs, trajectory=self._trj)
        result.__dict__.update(metadata)
        return result

    def to_shapely(self):
        """Return shapely object for area, bounds, etc. functions.

        Args:

        Returns:
          shapely.geometry.linestring.LineString -- Shapely shape.
          
        .. doctest::

            >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3]})
            >>> shape = df.traja.to_shapely()
            >>> shape.is_closed
            False

        """
        df = self._trj[['x', 'y']].dropna()
        coords = df.values
        tracks_data = {'type': 'LineString',
                       'coordinates': coords}
        tracks_shape = shape(tracks_data)
        return tracks_shape

    def calc_displacement(self, assign=True):
        """Calculate displacement between consecutive indices.

        Args:
          assign (bool, optional): Assign displacement to TrajaDataFrame (Default value = True)

        Returns:
          displacement (:class:`pandas.Series`): Displacement series.
          
        .. doctest::

            >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3]})
            >>> df.traja.calc_displacement()
            0         NaN
            1    1.414214
            2    1.414214
            dtype: float64

        """
        displacement = np.sqrt(np.power(self._trj.x.shift() - self._trj.x, 2) +
                               np.power(self._trj.y.shift() - self._trj.y, 2))

        # dx = self._trj.x.diff()
        # dy = self._trj.y.diff()
        if assign:
            self._trj = self._trj.assign(displacement=displacement)
        return displacement

    def calc_angle(self, assign=True):
        """Calculate angle between steps as a function of displacement w.r.t x axis.

        Args:
          assign (bool, optional): Assign displacement to TrajaDataFrame (Default value = True)

        Returns:
          angle (:class:`pandas.Series`): Angle series.
          
        .. doctest::

            >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3]})
            >>> df.traja.calc_angle()
            0     NaN
            1    45.0
            2    45.0
            dtype: float64

        """
        if not self._has_cols(['dx', 'displacement']):
            displacement = self.calc_displacement()
        else:
            displacement = self._trj.displacement

        angle = np.rad2deg(np.arccos(np.abs(self._trj.x.diff()) / displacement))
        if assign:
            self._trj['angle'] = angle
        return angle

    def scale(self, scale, spatial_units="m"):
        """Scale trajectory when converting, eg, from pixels to meters.

        Args:
          spatial_units(str., optional): Spatial units (eg, 'm') (Default value = "m")
          scale(float): Scale to convert coordinates

        .. doctest::

            >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3]})
            >>> df.traja.scale(0.1)
            >>> df
                 x    y
            0  0.0  0.1
            1  0.1  0.2
            2  0.2  0.3

        """
        self._trj[['x', 'y']] *= scale
        self._trj.__dict__['spatial_units'] = spatial_units


    def _transfer_metavars(self, df):
        for attr in self._trj._metadata:
            df.__dict__[attr] = getattr(self._trj, attr, None)
        return df


    def rediscretize(self, R):
        """Resample a trajectory to a constant step length. R is rediscretized step length.

        Args:
          R (float): Rediscretized step length (eg, 0.02)

        Returns:
          Rediscretized coordinates.
          
        .. note::
          
            Based on the appendix in Bovet and Benhamou, (1988) and Jim McLean's
            `trajr <https://github.com/JimMcL/trajr>`_ implementation.
          
        .. doctest::

            >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3]})
            >>> df.traja.rediscretize(1.)
                      x         y
            0  0.000000  1.000000
            1  0.707107  1.707107
            2  1.414214  2.414214

        """
        rt = self._rediscretize_points(R)

        if len(rt) < 2:
            raise RuntimeError(f"Step length {R} is too large for path (path length {len(self._trj)})")
        rt = traja.from_xy(rt)
        self._transfer_metavars(rt)
        return rt

    def _rediscretize_points(self, R):
        """Helper function for :meth:`~traja.utils.rediscretize`.

        Args:
          R (float): Rediscretized step length (eg, 0.02)

        Returns:
          result (:class:`numpy.ndarray`): Rediscretized coordinates

        """
        # TODO: Implement with complex numbers
        points = self._trj[['x', 'y']].dropna().values.astype('float64')
        n_points = len(points)
        result = np.empty((128, 2))
        p0 = points[0]
        result[0] = p0
        step_nr = 0
        candidate_start = 1 # running index of candidate

        while candidate_start <= n_points:
            # Find the first point `curr_ind` for which |points[curr_ind] - p_0| >= R
            curr_ind = np.NaN
            for i in range(candidate_start, n_points):  # range of search space for next point
                d = np.linalg.norm(points[i] - result[step_nr])
                if d >= R:
                    curr_ind = i  # curr_ind is in [candidate, n_points)
                    break
            if np.isnan(curr_ind):
                # End of path
                break

            # The next point may lie on the same segment
            candidate_start = curr_ind

            # The next point lies on the segment p[k-1], p[k]
            curr_result_x = result[step_nr][0]
            prev_x = points[curr_ind - 1, 0]
            curr_result_y = result[step_nr][1]
            prev_y = points[curr_ind - 1, 1]

            # a = 1 if points[k, 0] <= xk_1 else 0
            lambda_ = np.arctan2(points[curr_ind, 1] - prev_y, points[curr_ind, 0] - prev_x) # angle
            cos_l = np.cos(lambda_)
            sin_l = np.sin(lambda_)
            U = (curr_result_x - prev_x) * cos_l + (curr_result_y - prev_y) * sin_l
            V = (curr_result_y - prev_y) * cos_l - (curr_result_x - prev_x) * sin_l

            # Compute distance H between (X_{i+1}, Y_{i+1}) and (x_{k-1}, y_{k-1})
            H = U + np.sqrt(abs(R ** 2 - V ** 2))
            XIp1 = H * cos_l + prev_x
            YIp1 = H * sin_l + prev_y

            # Increase array size progressively to make the code run (significantly) faster
            if len(result) <= step_nr + 1:
                result = np.concatenate((result, np.empty_like(result)))

            # Save the point
            result[step_nr + 1] = np.array([XIp1, YIp1])
            step_nr += 1

        # Truncate result
        result = result[:step_nr + 1]
        return result

    def calc_heading(self, assign=True):
        """Calculate trajectory heading.

        Args:
          assign (bool): (Default value = True)

        Returns:
            heading (:class:`pandas.Series`): heading as a `Series`

        ..doctest::

            >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3]})
            >>> df.traja.calc_heading()
            0     NaN
            1    45.0
            2    45.0
            Name: heading, dtype: float64

        """
        if not self._has_cols('angle'):
            angle = self.calc_angle(assign=True)
        else:
            angle = self._trj.angle
        df = self._trj
        dx = df.x.diff()
        dy = df.y.diff()
        # Get heading from angle
        mask = (dx > 0) & (dy >= 0)
        df.loc[mask, 'heading'] = angle[mask]
        mask = (dx >= 0) & (dy < 0)
        df.loc[mask, 'heading'] = -angle[mask]
        mask = (dx < 0) & (dy <= 0)
        df.loc[mask, 'heading'] = -(180 - angle[mask])
        mask = (dx <= 0) & (dy > 0)
        df.loc[mask, 'heading'] = (180 - angle[mask])
        if assign:
            self._trj['heading'] = df.heading
        return df.heading

    def calc_turn_angle(self, assign=True):
        """Calculate turn angle.


        Args:
          assign (bool):  (Default value = True)

        Returns:
            turn_angle (:class:`~pandas.Series`): Turn angle

        .. doctest::

            >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3]})
            >>> df.traja.calc_turn_angle()
            0    NaN
            1    NaN
            2    0.0
            Name: turn_angle, dtype: float64

        """
        if 'heading' not in self._trj:
            heading = self.calc_heading(assign=False)
        else:
            heading = self._trj.heading
        turn_angle = heading.diff().rename('turn_angle')
        # Correction for 360-degree angle range
        turn_angle[turn_angle >= 180] -= 360
        turn_angle[turn_angle < -180] += 360
        if assign:
            self._trj['turn_angle'] = turn_angle
        return turn_angle


#! /usr/local/env python3
import logging
import sys

import traja
import matplotlib as mpl
if 'sphinx' in sys.argv[0]:
    mpl.use('agg')
import pandas as pd
from pandas.api.types import is_numeric_dtype, is_datetime64_any_dtype

logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.ERROR)


class TrajaDataFrame(pd.DataFrame):
    """A TrajaDataFrame object is a subclass of pandas :class:`<~pandas.dataframe.DataFrame>`.

    Args:
      args: Typical arguments for pandas.DataFrame.

    Returns:
      traja.TrajaDataFrame -- TrajaDataFrame constructor.
      
      .. doctest::

      >>> traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3]})
         x  y
      0  0  1
      1  1  2
      2  2  3

    """

    _metadata = ['xlim', 'ylim', 'spatial_units', 'xlabel', 'ylabel', 'title', 'fps', 'time_units', 'time_col', 'id']

    def __init__(self, *args, **kwargs):
        # Allow setting metadata from constructor
        traja_kwargs = dict()
        for key in list(kwargs.keys()):
            for name in self._metadata:
                if key == name:
                    traja_kwargs[key] = kwargs.pop(key)
        super(TrajaDataFrame, self).__init__(*args, **kwargs)
        if len(args) == 1 and isinstance(args[0], TrajaDataFrame):
            args[0]._copy_attrs(self)
        for name, value in traja_kwargs.items():
            self.__dict__[name] = value

        # Initialize metadata like 'fps','spatial_units', etc.
        self._init_metadata()

    @property
    def _constructor(self):
        return TrajaDataFrame

    def _copy_attrs(self, df):
        for attr in self._metadata:
            df.__dict__[attr] = getattr(self, attr, None)

    def __finalize__(self, other, method=None, **kwargs):
        """propagate metadata from other to self """
        # merge operation: using metadata of the left object
        if method == 'merge':
            for name in self._metadata:
                object.__setattr__(self, name, getattr(other.left, name, None))
        # concat operation: using metadata of the first object
        elif method == 'concat':
            for name in self._metadata:
                object.__setattr__(self, name, getattr(other.objs[0], name, None))
        else:
            for name in self._metadata:
                object.__setattr__(self, name, getattr(other, name, None))

        # Requires 'time' column, so find one
        # time_col = self._get_time_col()
        # fps = kwargs.pop('fps',1)
        # assert isinstance(fps, (int, float)), f"{fps} is not a float or int"
        # if time_col is not None:
        #     self.rename(columns={time_col:'time'})
        # else:
        #     # Otherwise, create one from index and `fps`
        #     self['time'] = self.index
        #     self.time /= fps

        return self

    def _init_metadata(self):
        defaults = dict(fps=None,
                        spatial_units='m',
                        time_units='s')
        for name, value in defaults.items():
            if name not in self.__dict__:
                self.__dict__[name] = value

    def _get_time_col(self):
        time_cols = [col for col in self if 'time' in col.lower()]
        if time_cols:
            time_col = time_cols[0]
            if is_numeric_dtype(self[time_col]):
                return time_col
        else:
            return None

    def copy(self, deep=True):
        """Make a copy of this TrajaDataFrame object

        Args:
          deep(bool, optional): Make a deep copy, i.e. also copy data (Default value = True)

        Returns:
          TrajaDataFrame -- copy

        """
        data = self._data
        if deep:
            data = data.copy()
        return TrajaDataFrame(data).__finalize__(self)


# TODO: Replace with tests.
# class Debug():
#     """Debug only.
#     """
#
#     def __init__(self, n_coords=1000):
#         import glob
#         import traja
#         from traja import TrajaAccessor
#         # files = glob.glob('/Users/justinshenk/neurodata/data/raw_centroids_rev2/*')
#         # df = traja.read_file(files[10])
#         # df.traja.set(xlim=(-0.06, 0.06),
#         #              ylim=(-0.13, 0.13),
#         #              xlabel=("x (m)"),
#         #              ylabel=("y (m)"),
#         #              title="Cage trajectory")
#         # FIXME: Function below takes forerver (or doesn't complete)
#         basepath = os.path.dirname(traja.__file__)
#         filepath = os.path.join(basepath, 'test', 'test_data', '3527.csv')
#         df = traja.read_file(filepath)
#         df.traja.plot()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

import os
from os import path
import re

here = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    with open(os.path.join(here, *parts), 'r', encoding='utf8') as fp:
        return fp.read()

# Get package version
def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

requirements = ['matplotlib','pandas','numpy','shapely','psutil', 'scipy']

this_dir = path.abspath(path.dirname(__file__))
with open(os.path.join(this_dir, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='traja',
    version=find_version('traja','__init__.py'),
    description='Traja is a trajectory analysis and visualization tool',
    url='https://github.com/justinshenk/traja',
    author='Justin Shenk',
    author_email='shenkjustin@gmail.com',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=requirements,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Libraries',
    ],
    python_requires='!= 3.0.*, != 3.1.*',
    packages=find_packages(),
    include_package_data=True,
    license='MIT',
    keywords='trajectory analysis',
    zip_safe=False,
)
